#include "scene.h"

Scene *scene;

#include <iostream>

using namespace std;


Scene::Scene() {

  shadowMap = 0;
  lightCamera = 0;

  gui = new GUI();
  gui->show();

  int u0 = 20;
  int v0 = 50;
  int sci = 2;
  int w = sci*320;
  int h = sci*240;
  fb = new FrameBuffer(u0, v0, w, h);
  fb->label("SW framebuffer");
  fb->show();
  gui->uiw->position(fb->w+20+20, 50);

  float hfov = 45.0f;
  ppc = new PPC(fb->w, fb->h, hfov);

  cube = new V3[8];
  float hl = 10.0f;
  float zpos = -100.0f;
  cube[0] = V3(-hl, hl, hl+zpos);
  cube[1] = V3(-hl, -hl, hl+zpos);
  cube[2] = V3(hl, -hl, hl+zpos);
  cube[3] = V3(hl, hl, hl+zpos);
  cube[4] = V3(-hl, hl, -hl+zpos);
  cube[5] = V3(-hl, -hl, -hl+zpos);
  cube[6] = V3(hl, -hl, -hl+zpos);
  cube[7] = V3(hl, hl, -hl+zpos);

  tmeshesN = 3;
  tmeshes = new TMesh[tmeshesN];
  tmeshes[0].Load("geometry/teapot1K.bin");
  tmeshes[0].Scale(100.0f);
  tmeshes[0].SetCenter(V3(-20.0f, 0.0f, -200.0f));
  tmeshes[0].SetColors(V3(1.0f, 0.0f, 0.0f));
  AABB aabb = tmeshes[0].GetAABB();
  tmeshes[1].SetBaseRectangle(
    V3(0.0f, aabb[0][1]-1.0f, -276.0f), 150.0f, 75.0f);
  tmeshes[2].SetBaseRectangle(
    V3(0.0f, aabb[0][1]-1.0f, -200.0f), 150.0f, 75.0f);

  lightPoint = V3(-50.0f, 50.0f, -200.0f);

  Render();

}

void Scene::Render() {


  fb->FrameSetup();

//  UpdateShadowMap();

  tmeshes[1].RenderFilled(fb, ppc, lightPoint, 0, 0, false);
  tmeshes[2].RenderFilled(fb, ppc, lightPoint, 0, 0, true);

//  fb->DrawPoint3D(lightPoint, 7, 0xFF000000, ppc);

#if 0
  int psize = 7;
  unsigned int pcol = 0xFF000000;
  for (int i = 0; i < 8; i++) {
    fb->DrawPoint3D(cube[i], psize, pcol, ppc);
  }

  for (int i = 0; i < 4; i++) {
    int _i = (i+1) % 4;
    fb->DrawSegment3D(cube[i], cube[_i],
      0xFF0000FF, ppc);
    fb->DrawSegment3D(cube[4+i], cube[4+_i],
      0xFF0000FF, ppc);
    fb->DrawSegment3D(cube[i], cube[4+i],
      0xFF0000FF, ppc);
  }
#endif

  fb->redraw();

}

void Scene::UpdateShadowMap() {

  // set light camera
  if (!lightCamera)
    lightCamera = new PPC(256, 256, 30.0f);
  lightCamera->C = lightPoint;
  float f = lightCamera->GetF();
  V3 vd = (tmeshes[0].GetCenter() - lightPoint).Normalized();
  lightCamera->a = 
    (vd^V3(0.0f, 1.0f, 0.0f)).Normalized() * lightCamera->a.Length();
  lightCamera->b = (vd ^ lightCamera->a).Normalized() *
    lightCamera->b.Length();
  lightCamera->c = vd*f 
    - lightCamera->a*((float)lightCamera->w)/2.0f
    - lightCamera->b*((float)lightCamera->h)/2.0f;

  AABB aabb = tmeshes[0].GetAABB();
  lightCamera->ZoomToAABB(aabb);

  // render scene with light camera
  if (!shadowMap) {
    shadowMap = new FrameBuffer(20, 50 + fb->h + 50,
      lightCamera->w, lightCamera->h);
    shadowMap->label("Shadow Map");
    shadowMap->show();
  }
  shadowMap->FrameSetup();
  for (int i = 0; i < tmeshesN; i++) {
    tmeshes[i].RenderFilled(
      shadowMap, lightCamera, 
      lightPoint, 0, 0, false);
  }
  shadowMap->redraw();

}

void Scene::DBG() {

  fb->Save("mdybg/fb.tif");
  return;


  {

    V3 l0(-50.0f, 50.0f, -200.0f);
    V3 l1(+50.0f, 50.0f, -200.0f);
    int framesN = 100;
    for (int i = 0; i < framesN; i++) {
      float frac = (float) i / (float) (framesN-1);
      lightPoint = l0 + (l1-l0)*frac;
      Render();
      Fl::check();
    }

    return;

  }


  {
    PPC ppc0 = *ppc;
    PPC ppcn = *ppc;
//    ppcn.TranslateForward(20.0f);
    ppcn.TranslateRight(200.0f);
    ppcn.RotateAboutAxisThroughC(ppcn.b.Normalized()*-1.0f, 50.0f);
    int n = 10;
    for (int i = 0; i < n; i++) {
      *ppc = ppc0.Interpolate(&ppcn, i, n);
      Render();
      Fl::check();
    }
    return;
  }

  {

    V3 P(-30.0f, 50.0f, -100.0f);
    V3 projP;
    bool projOK = ppc->Project(P, projP);
    cerr << projP << endl;
    return;

  }

  {

    int cw = 40;
    unsigned col0 = 0xFFAAAAAA;
    unsigned col1 = 0xFFFFFFFF;
    fb->MakeChecker(cw, col0, col1);
    fb->redraw();

  }

  {

    V3 v0(1.0f, 1.0f, 1.0f);
    V3 v1(2.0f, 2.0f, 2.0f);
    V3 v2;
    v2 = v0 + v1;
    v0[0] = 2.0f;
    cerr << "v0: " << v0 << endl;
    cerr << "v1: " << v1 << endl;
    cerr << "v2: " << v2 << endl;
    cerr << endl;

    M3x3 m;
    m[0] = v0;
    m[1] = v1;
    m[2] = v2;
    m[2][2] = 13.0f;
    cerr << m[0] << endl << m[1] << endl << m[2] << endl;


    return;

  }



}

void Scene::NewButton() {

  cerr << "INFO: pressed New Button" << endl;

}


void Scene::View0() {

  ppc0 = *ppc;

}

void Scene::View1() {

  ppc1 = *ppc;

}

void Scene::Int01() {

    int n = 100;
    for (int i = 0; i < n; i++) {
      *ppc = ppc0.Interpolate(&ppc1, i, n);
      Render();
      Fl::check();
    }

}

void Scene::Int10() {

    int n = 100;
    for (int i = 0; i < n; i++) {
      *ppc = ppc1.Interpolate(&ppc0, i, n);
      Render();
      Fl::check();
    }

}


void Scene::SaveView() {

  ppc->Save("mydbg/view.txt");

}

void Scene::LoadView() {

  ppc->Load("mydbg/view.txt");
  Render();

}